(************************************************************************)
(*         *      The Rocq Prover / The Rocq Development Team           *)
(*  v      *         Copyright INRIA, CNRS and contributors             *)
(* <O___,, * (see version control and CREDITS file for authors & dates) *)
(*   \VV/  **************************************************************)
(*    //   *    This file is distributed under the terms of the         *)
(*         *     GNU Lesser General Public License Version 2.1          *)
(*         *     (see LICENSE file for the text of the license)         *)
(************************************************************************)

From Stdlib Require Import Rbase.
From Stdlib Require Import Rfunctions.
From Stdlib Require Import Rtrigo.
From Stdlib Require Import SeqSeries.
From Stdlib Require Export Ranalysis1.
From Stdlib Require Export Ranalysis2.
From Stdlib Require Export Ranalysis3.
From Stdlib Require Export Rtopology.
From Stdlib Require Export MVT.
From Stdlib Require Export PSeries_reg.
From Stdlib Require Export Exp_prop.
From Stdlib Require Export Rtrigo_reg.
From Stdlib Require Export Rsqrt_def.
From Stdlib Require Export R_sqrt.
From Stdlib Require Export Rtrigo_calc.
From Stdlib Require Export Rgeom.
From Stdlib Require Export Sqrt_reg.
From Stdlib Require Export Ranalysis4.
From Stdlib Require Export Rpower.
From Stdlib Require Export Ranalysis_reg.
