(************************************************************************)
(*         *      The Rocq Prover / The Rocq Development Team           *)
(*  v      *         Copyright INRIA, CNRS and contributors             *)
(* <O___,, * (see version control and CREDITS file for authors & dates) *)
(*   \VV/  **************************************************************)
(*    //   *    This file is distributed under the terms of the         *)
(*         *     GNU Lesser General Public License Version 2.1          *)
(*         *     (see LICENSE file for the text of the license)         *)
(************************************************************************)


From Stdlib Require Export OrderedType OrderedTypeEx OrderedTypeAlt.
From Stdlib Require Export DecidableType DecidableTypeEx.
From Stdlib Require Export FMapInterface.
From Stdlib Require Export FMapPositive.
From Stdlib Require Export FMapFacts.
From Stdlib Require Export FMapWeakList.
From Stdlib Require Export FMapList.
