Set Warnings "-deprecated-library-file,-warn-library-file,-notation-incompatible-prefix,-notation-overridden,-overwriting-delimiting-key".
From Stdlib Require Export Arith.Arith.
From Stdlib Require Export Arith.Arith_base.
From Stdlib Require Export Arith.Between.
From Stdlib Require Export Arith.Bool_nat.
From Stdlib Require Export Arith.Cantor.
From Stdlib Require Export Arith.Compare.
From Stdlib Require Export Arith.Compare_dec.
From Stdlib Require Export Arith.EqNat.
From Stdlib Require Export Arith.Euclid.
From Stdlib Require Export Arith.Factorial.
From Stdlib Require Export Arith.PeanoNat.
From Stdlib Require Export Arith.Peano_dec.
From Stdlib Require Export Arith.Wf_nat.
From Stdlib Require Export Arith.Zerob.
From Stdlib Require Export Array.ArrayAxioms.
From Stdlib Require Export Array.PArray.
From Stdlib Require Export Array.PrimArray.
From Stdlib Require Export BinNums.IntDef.
From Stdlib Require Export BinNums.NatDef.
From Stdlib Require Export BinNums.PosDef.
From Stdlib Require Export Bool.Bool.
From Stdlib Require Export Bool.BoolEq.
From Stdlib Require Export Bool.DecBool.
From Stdlib Require Export Bool.IfProp.
From Stdlib Require Export Classes.CEquivalence.
From Stdlib Require Export Classes.CMorphisms.
From Stdlib Require Export Classes.CRelationClasses.
From Stdlib Require Export Classes.DecidableClass.
From Stdlib Require Export Classes.EquivDec.
From Stdlib Require Export Classes.Equivalence.
From Stdlib Require Export Classes.Init.
From Stdlib Require Export Classes.Morphisms.
From Stdlib Require Export Classes.Morphisms_Prop.
From Stdlib Require Export Classes.Morphisms_Relations.
From Stdlib Require Export Classes.RelationClasses.
From Stdlib Require Export Classes.RelationPairs.
From Stdlib Require Export Classes.SetoidClass.
From Stdlib Require Export Classes.SetoidDec.
From Stdlib Require Export Classes.SetoidTactics.
From Stdlib Require Export Compat.AdmitAxiom.
From Stdlib Require Export Compat.Coq818.
From Stdlib Require Export Compat.Coq819.
From Stdlib Require Export Compat.Coq820.
From Stdlib Require Export Compat.Stdlib818.
From Stdlib Require Export FSets.FMapAVL.
From Stdlib Require Export FSets.FMapFacts.
From Stdlib Require Export FSets.FMapFullAVL.
From Stdlib Require Export FSets.FMapInterface.
From Stdlib Require Export FSets.FMapList.
From Stdlib Require Export FSets.FMapPositive.
From Stdlib Require Export FSets.FMapWeakList.
From Stdlib Require Export FSets.FMaps.
From Stdlib Require Export FSets.FSetAVL.
From Stdlib Require Export FSets.FSetBridge.
From Stdlib Require Export FSets.FSetCompat.
From Stdlib Require Export FSets.FSetDecide.
From Stdlib Require Export FSets.FSetEqProperties.
From Stdlib Require Export FSets.FSetFacts.
From Stdlib Require Export FSets.FSetInterface.
From Stdlib Require Export FSets.FSetList.
From Stdlib Require Export FSets.FSetPositive.
From Stdlib Require Export FSets.FSetProperties.
From Stdlib Require Export FSets.FSetToFiniteSet.
From Stdlib Require Export FSets.FSetWeakList.
From Stdlib Require Export FSets.FSets.
From Stdlib Require Export Floats.FloatAxioms.
From Stdlib Require Export Floats.FloatClass.
From Stdlib Require Export Floats.FloatLemmas.
From Stdlib Require Export Floats.FloatOps.
From Stdlib Require Export Floats.Floats.
From Stdlib Require Export Floats.PrimFloat.
From Stdlib Require Export Floats.SpecFloat.
From Stdlib Require Export Init.Byte.
From Stdlib Require Export Init.Datatypes.
From Stdlib Require Export Init.Decimal.
From Stdlib Require Export Init.Hexadecimal.
From Stdlib Require Export Init.Logic.
From Stdlib Require Export Init.Ltac.
From Stdlib Require Export Init.Nat.
From Stdlib Require Export Init.Notations.
From Stdlib Require Export Init.Number.
From Stdlib Require Export Init.Peano.
From Stdlib Require Export Init.Prelude.
From Stdlib Require Export Init.Specif.
From Stdlib Require Export Init.Sumbool.
From Stdlib Require Export Init.Tactics.
From Stdlib Require Export Init.Tauto.
From Stdlib Require Export Init.Wf.
From Stdlib Require Export Lists.Finite.
From Stdlib Require Export Lists.List.
From Stdlib Require Export Lists.ListDec.
From Stdlib Require Export Lists.ListDef.
From Stdlib Require Export Lists.ListSet.
From Stdlib Require Export Lists.ListTactics.
From Stdlib Require Export Logic.Adjointification.
From Stdlib Require Export Logic.Berardi.
From Stdlib Require Export Logic.ChoiceFacts.
From Stdlib Require Export Logic.Classical.
From Stdlib Require Export Logic.ClassicalChoice.
From Stdlib Require Export Logic.ClassicalDescription.
From Stdlib Require Export Logic.ClassicalEpsilon.
From Stdlib Require Export Logic.ClassicalFacts.
From Stdlib Require Export Logic.ClassicalUniqueChoice.
From Stdlib Require Export Logic.Classical_Pred_Type.
From Stdlib Require Export Logic.Classical_Prop.
From Stdlib Require Export Logic.ConstructiveEpsilon.
From Stdlib Require Export Logic.Decidable.
From Stdlib Require Export Logic.Description.
From Stdlib Require Export Logic.Diaconescu.
From Stdlib Require Export Logic.Epsilon.
From Stdlib Require Export Logic.Eqdep.
From Stdlib Require Export Logic.EqdepFacts.
From Stdlib Require Export Logic.Eqdep_dec.
From Stdlib Require Export Logic.ExtensionalFunctionRepresentative.
From Stdlib Require Export Logic.ExtensionalityFacts.
From Stdlib Require Export Logic.FunctionalExtensionality.
From Stdlib Require Export Logic.HLevels.
From Stdlib Require Export Logic.HLevelsBase.
From Stdlib Require Export Logic.Hurkens.
From Stdlib Require Export Logic.IndefiniteDescription.
From Stdlib Require Export Logic.JMeq.
From Stdlib Require Export Logic.ProofIrrelevance.
From Stdlib Require Export Logic.ProofIrrelevanceFacts.
From Stdlib Require Export Logic.PropExtensionality.
From Stdlib Require Export Logic.PropExtensionalityFacts.
From Stdlib Require Export Logic.PropFacts.
From Stdlib Require Export Logic.RelationalChoice.
From Stdlib Require Export Logic.SetIsType.
From Stdlib Require Export Logic.SetoidChoice.
From Stdlib Require Export Logic.StrictProp.
From Stdlib Require Export Logic.WKL.
From Stdlib Require Export Logic.WeakFan.
From Stdlib Require Export MSets.MSetAVL.
From Stdlib Require Export MSets.MSetDecide.
From Stdlib Require Export MSets.MSetEqProperties.
From Stdlib Require Export MSets.MSetFacts.
From Stdlib Require Export MSets.MSetGenTree.
From Stdlib Require Export MSets.MSetInterface.
From Stdlib Require Export MSets.MSetList.
From Stdlib Require Export MSets.MSetPositive.
From Stdlib Require Export MSets.MSetProperties.
From Stdlib Require Export MSets.MSetRBT.
From Stdlib Require Export MSets.MSetToFiniteSet.
From Stdlib Require Export MSets.MSetWeakList.
From Stdlib Require Export MSets.MSets.
From Stdlib Require Export NArith.BinNat.
From Stdlib Require Export NArith.BinNatDef.
From Stdlib Require Export NArith.NArith.
From Stdlib Require Export NArith.NArith_base.
From Stdlib Require Export NArith.Ndec.
From Stdlib Require Export NArith.Ndiv_def.
From Stdlib Require Export NArith.Ngcd_def.
From Stdlib Require Export NArith.Nnat.
From Stdlib Require Export NArith.Nsqrt_def.
From Stdlib Require Export Numbers.AltBinNotations.
From Stdlib Require Export Numbers.BinNums.
From Stdlib Require Export Numbers.Cyclic.Abstract.CyclicAxioms.
From Stdlib Require Export Numbers.Cyclic.Abstract.DoubleType.
From Stdlib Require Export Numbers.Cyclic.Abstract.NZCyclic.
From Stdlib Require Export Numbers.Cyclic.Int63.CarryType.
From Stdlib Require Export Numbers.Cyclic.Int63.Cyclic63.
From Stdlib Require Export Numbers.Cyclic.Int63.PrimInt63.
From Stdlib Require Export Numbers.Cyclic.Int63.Ring63.
From Stdlib Require Export Numbers.Cyclic.Int63.Sint63.
From Stdlib Require Export Numbers.Cyclic.Int63.Sint63Axioms.
From Stdlib Require Export Numbers.Cyclic.Int63.Uint63.
From Stdlib Require Export Numbers.Cyclic.Int63.Uint63Axioms.
From Stdlib Require Export Numbers.DecimalFacts.
From Stdlib Require Export Numbers.DecimalN.
From Stdlib Require Export Numbers.DecimalNat.
From Stdlib Require Export Numbers.DecimalPos.
From Stdlib Require Export Numbers.DecimalQ.
From Stdlib Require Export Numbers.DecimalR.
From Stdlib Require Export Numbers.DecimalString.
From Stdlib Require Export Numbers.DecimalZ.
From Stdlib Require Export Numbers.HexadecimalFacts.
From Stdlib Require Export Numbers.HexadecimalN.
From Stdlib Require Export Numbers.HexadecimalNat.
From Stdlib Require Export Numbers.HexadecimalPos.
From Stdlib Require Export Numbers.HexadecimalQ.
From Stdlib Require Export Numbers.HexadecimalR.
From Stdlib Require Export Numbers.HexadecimalString.
From Stdlib Require Export Numbers.HexadecimalZ.
From Stdlib Require Export Numbers.Integer.Abstract.ZAdd.
From Stdlib Require Export Numbers.Integer.Abstract.ZAddOrder.
From Stdlib Require Export Numbers.Integer.Abstract.ZAxioms.
From Stdlib Require Export Numbers.Integer.Abstract.ZBase.
From Stdlib Require Export Numbers.Integer.Abstract.ZBits.
From Stdlib Require Export Numbers.Integer.Abstract.ZDivEucl.
From Stdlib Require Export Numbers.Integer.Abstract.ZDivFloor.
From Stdlib Require Export Numbers.Integer.Abstract.ZDivTrunc.
From Stdlib Require Export Numbers.Integer.Abstract.ZGcd.
From Stdlib Require Export Numbers.Integer.Abstract.ZLcm.
From Stdlib Require Export Numbers.Integer.Abstract.ZLt.
From Stdlib Require Export Numbers.Integer.Abstract.ZMaxMin.
From Stdlib Require Export Numbers.Integer.Abstract.ZMul.
From Stdlib Require Export Numbers.Integer.Abstract.ZMulOrder.
From Stdlib Require Export Numbers.Integer.Abstract.ZParity.
From Stdlib Require Export Numbers.Integer.Abstract.ZPow.
From Stdlib Require Export Numbers.Integer.Abstract.ZProperties.
From Stdlib Require Export Numbers.Integer.Abstract.ZSgnAbs.
From Stdlib Require Export Numbers.Integer.Binary.ZBinary.
From Stdlib Require Export Numbers.Integer.NatPairs.ZNatPairs.
From Stdlib Require Export Numbers.NaryFunctions.
From Stdlib Require Export Numbers.NatInt.NZAdd.
From Stdlib Require Export Numbers.NatInt.NZAddOrder.
From Stdlib Require Export Numbers.NatInt.NZAxioms.
From Stdlib Require Export Numbers.NatInt.NZBase.
From Stdlib Require Export Numbers.NatInt.NZBits.
From Stdlib Require Export Numbers.NatInt.NZDiv.
From Stdlib Require Export Numbers.NatInt.NZDomain.
From Stdlib Require Export Numbers.NatInt.NZGcd.
From Stdlib Require Export Numbers.NatInt.NZLog.
From Stdlib Require Export Numbers.NatInt.NZMul.
From Stdlib Require Export Numbers.NatInt.NZMulOrder.
From Stdlib Require Export Numbers.NatInt.NZOrder.
From Stdlib Require Export Numbers.NatInt.NZParity.
From Stdlib Require Export Numbers.NatInt.NZPow.
From Stdlib Require Export Numbers.NatInt.NZProperties.
From Stdlib Require Export Numbers.NatInt.NZSqrt.
From Stdlib Require Export Numbers.Natural.Abstract.NAdd.
From Stdlib Require Export Numbers.Natural.Abstract.NAddOrder.
From Stdlib Require Export Numbers.Natural.Abstract.NAxioms.
From Stdlib Require Export Numbers.Natural.Abstract.NBase.
From Stdlib Require Export Numbers.Natural.Abstract.NBits.
From Stdlib Require Export Numbers.Natural.Abstract.NDefOps.
From Stdlib Require Export Numbers.Natural.Abstract.NDiv.
From Stdlib Require Export Numbers.Natural.Abstract.NDiv0.
From Stdlib Require Export Numbers.Natural.Abstract.NGcd.
From Stdlib Require Export Numbers.Natural.Abstract.NIso.
From Stdlib Require Export Numbers.Natural.Abstract.NLcm.
From Stdlib Require Export Numbers.Natural.Abstract.NLcm0.
From Stdlib Require Export Numbers.Natural.Abstract.NLog.
From Stdlib Require Export Numbers.Natural.Abstract.NMaxMin.
From Stdlib Require Export Numbers.Natural.Abstract.NMulOrder.
From Stdlib Require Export Numbers.Natural.Abstract.NOrder.
From Stdlib Require Export Numbers.Natural.Abstract.NParity.
From Stdlib Require Export Numbers.Natural.Abstract.NPow.
From Stdlib Require Export Numbers.Natural.Abstract.NProperties.
From Stdlib Require Export Numbers.Natural.Abstract.NSqrt.
From Stdlib Require Export Numbers.Natural.Abstract.NStrongRec.
From Stdlib Require Export Numbers.Natural.Abstract.NSub.
From Stdlib Require Export Numbers.Natural.Binary.NBinary.
From Stdlib Require Export Numbers.NumPrelude.
From Stdlib Require Export PArith.BinPos.
From Stdlib Require Export PArith.BinPosDef.
From Stdlib Require Export PArith.PArith.
From Stdlib Require Export PArith.POrderedType.
From Stdlib Require Export PArith.Pnat.
From Stdlib Require Export Program.Basics.
From Stdlib Require Export Program.Combinators.
From Stdlib Require Export Program.Equality.
From Stdlib Require Export Program.Program.
From Stdlib Require Export Program.Subset.
From Stdlib Require Export Program.Syntax.
From Stdlib Require Export Program.Tactics.
From Stdlib Require Export Program.Utils.
From Stdlib Require Export Program.Wf.
From Stdlib Require Export Program.WfExtensionality.
From Stdlib Require Export QArith.QArith.
From Stdlib Require Export QArith.QArith_base.
From Stdlib Require Export QArith.QNsatz.
From Stdlib Require Export QArith.QOrderedType.
From Stdlib Require Export QArith.Qabs.
From Stdlib Require Export QArith.Qcabs.
From Stdlib Require Export QArith.Qcanon.
From Stdlib Require Export QArith.Qfield.
From Stdlib Require Export QArith.Qminmax.
From Stdlib Require Export QArith.Qpower.
From Stdlib Require Export QArith.Qreduction.
From Stdlib Require Export QArith.Qring.
From Stdlib Require Export QArith.Qround.
From Stdlib Require Export Reals.Abstract.ConstructiveAbs.
From Stdlib Require Export Reals.Abstract.ConstructiveLUB.
From Stdlib Require Export Reals.Abstract.ConstructiveLimits.
From Stdlib Require Export Reals.Abstract.ConstructiveMinMax.
From Stdlib Require Export Reals.Abstract.ConstructivePower.
From Stdlib Require Export Reals.Abstract.ConstructiveReals.
From Stdlib Require Export Reals.Abstract.ConstructiveRealsMorphisms.
From Stdlib Require Export Reals.Abstract.ConstructiveSum.
From Stdlib Require Export Reals.Alembert.
From Stdlib Require Export Reals.AltSeries.
From Stdlib Require Export Reals.ArithProp.
From Stdlib Require Export Reals.Binomial.
From Stdlib Require Export Reals.Cauchy.ConstructiveCauchyAbs.
From Stdlib Require Export Reals.Cauchy.ConstructiveCauchyReals.
From Stdlib Require Export Reals.Cauchy.ConstructiveCauchyRealsMult.
From Stdlib Require Export Reals.Cauchy.ConstructiveExtra.
From Stdlib Require Export Reals.Cauchy.ConstructiveRcomplete.
From Stdlib Require Export Reals.Cauchy.PosExtra.
From Stdlib Require Export Reals.Cauchy.QExtra.
From Stdlib Require Export Reals.Cauchy_prod.
From Stdlib Require Export Reals.ClassicalConstructiveReals.
From Stdlib Require Export Reals.ClassicalDedekindReals.
From Stdlib Require Export Reals.Cos_plus.
From Stdlib Require Export Reals.Cos_rel.
From Stdlib Require Export Reals.DiscrR.
From Stdlib Require Export Reals.Exp_prop.
From Stdlib Require Export Reals.Integration.
From Stdlib Require Export Reals.MVT.
From Stdlib Require Export Reals.Machin.
From Stdlib Require Export Reals.NewtonInt.
From Stdlib Require Export Reals.Nsatz.
From Stdlib Require Export Reals.PSeries_reg.
From Stdlib Require Export Reals.PartSum.
From Stdlib Require Export Reals.Qreals.
From Stdlib Require Export Reals.RIneq.
From Stdlib Require Export Reals.RList.
From Stdlib Require Export Reals.RNsatz.
From Stdlib Require Export Reals.ROrderedType.
From Stdlib Require Export Reals.R_Ifp.
From Stdlib Require Export Reals.R_sqr.
From Stdlib Require Export Reals.R_sqrt.
From Stdlib Require Export Reals.Ranalysis.
From Stdlib Require Export Reals.Ranalysis1.
From Stdlib Require Export Reals.Ranalysis2.
From Stdlib Require Export Reals.Ranalysis3.
From Stdlib Require Export Reals.Ranalysis4.
From Stdlib Require Export Reals.Ranalysis5.
From Stdlib Require Export Reals.Ranalysis_reg.
From Stdlib Require Export Reals.Ratan.
From Stdlib Require Export Reals.Raxioms.
From Stdlib Require Export Reals.Rbase.
From Stdlib Require Export Reals.Rbasic_fun.
From Stdlib Require Export Reals.Rcomplete.
From Stdlib Require Export Reals.Rdefinitions.
From Stdlib Require Export Reals.Rderiv.
From Stdlib Require Export Reals.Reals.
From Stdlib Require Export Reals.Rfunctions.
From Stdlib Require Export Reals.Rgeom.
From Stdlib Require Export Reals.RiemannInt.
From Stdlib Require Export Reals.RiemannInt_SF.
From Stdlib Require Export Reals.Rlimit.
From Stdlib Require Export Reals.Rlogic.
From Stdlib Require Export Reals.Rminmax.
From Stdlib Require Export Reals.Rpow_def.
From Stdlib Require Export Reals.Rpower.
From Stdlib Require Export Reals.Rprod.
From Stdlib Require Export Reals.Rregisternames.
From Stdlib Require Export Reals.Rseries.
From Stdlib Require Export Reals.Rsigma.
From Stdlib Require Export Reals.Rsqrt_def.
From Stdlib Require Export Reals.Rtopology.
From Stdlib Require Export Reals.Rtrigo.
From Stdlib Require Export Reals.Rtrigo1.
From Stdlib Require Export Reals.Rtrigo_alt.
From Stdlib Require Export Reals.Rtrigo_calc.
From Stdlib Require Export Reals.Rtrigo_def.
From Stdlib Require Export Reals.Rtrigo_facts.
From Stdlib Require Export Reals.Rtrigo_fun.
From Stdlib Require Export Reals.Rtrigo_reg.
From Stdlib Require Export Reals.Runcountable.
From Stdlib Require Export Reals.SeqProp.
From Stdlib Require Export Reals.SeqSeries.
From Stdlib Require Export Reals.SplitAbsolu.
From Stdlib Require Export Reals.SplitRmult.
From Stdlib Require Export Reals.Sqrt_reg.
From Stdlib Require Export Reals.Zfloor.
From Stdlib Require Export Relations.Operators_Properties.
From Stdlib Require Export Relations.Relation_Definitions.
From Stdlib Require Export Relations.Relation_Operators.
From Stdlib Require Export Relations.Relations.
From Stdlib Require Export Setoids.Setoid.
From Stdlib Require Export Sets.Classical_sets.
From Stdlib Require Export Sets.Constructive_sets.
From Stdlib Require Export Sets.Cpo.
From Stdlib Require Export Sets.Ensembles.
From Stdlib Require Export Sets.Finite_sets.
From Stdlib Require Export Sets.Finite_sets_facts.
From Stdlib Require Export Sets.Image.
From Stdlib Require Export Sets.Infinite_sets.
From Stdlib Require Export Sets.Integers.
From Stdlib Require Export Sets.Multiset.
From Stdlib Require Export Sets.Partial_Order.
From Stdlib Require Export Sets.Permut.
From Stdlib Require Export Sets.Powerset.
From Stdlib Require Export Sets.Powerset_Classical_facts.
From Stdlib Require Export Sets.Powerset_facts.
From Stdlib Require Export Sets.Relations_1.
From Stdlib Require Export Sets.Relations_1_facts.
From Stdlib Require Export Sets.Relations_2.
From Stdlib Require Export Sets.Relations_2_facts.
From Stdlib Require Export Sets.Relations_3.
From Stdlib Require Export Sets.Relations_3_facts.
From Stdlib Require Export Sets.Uniset.
From Stdlib Require Export Sorting.CPermutation.
From Stdlib Require Export Sorting.Heap.
From Stdlib Require Export Sorting.Mergesort.
From Stdlib Require Export Sorting.PermutEq.
From Stdlib Require Export Sorting.PermutSetoid.
From Stdlib Require Export Sorting.Permutation.
From Stdlib Require Export Sorting.SetoidList.
From Stdlib Require Export Sorting.SetoidPermutation.
From Stdlib Require Export Sorting.Sorted.
From Stdlib Require Export Sorting.Sorting.
From Stdlib Require Export Streams.StreamMemo.
From Stdlib Require Export Streams.Streams.
From Stdlib Require Export Strings.Ascii.
From Stdlib Require Export Strings.BinaryString.
From Stdlib Require Export Strings.Byte.
From Stdlib Require Export Strings.HexString.
From Stdlib Require Export Strings.OctalString.
From Stdlib Require Export Strings.PString.
From Stdlib Require Export Strings.PrimString.
From Stdlib Require Export Strings.PrimStringAxioms.
From Stdlib Require Export Strings.String.
From Stdlib Require Export Structures.BoolOrder.
From Stdlib Require Export Structures.DecidableType.
From Stdlib Require Export Structures.DecidableTypeEx.
From Stdlib Require Export Structures.Equalities.
From Stdlib Require Export Structures.EqualitiesFacts.
From Stdlib Require Export Structures.GenericMinMax.
From Stdlib Require Export Structures.OrderedType.
From Stdlib Require Export Structures.OrderedTypeAlt.
From Stdlib Require Export Structures.OrderedTypeEx.
From Stdlib Require Export Structures.Orders.
From Stdlib Require Export Structures.OrdersAlt.
From Stdlib Require Export Structures.OrdersEx.
From Stdlib Require Export Structures.OrdersFacts.
From Stdlib Require Export Structures.OrdersLists.
From Stdlib Require Export Structures.OrdersTac.
From Stdlib Require Export Unicode.Utf8.
From Stdlib Require Export Unicode.Utf8_core.
From Stdlib Require Export Vectors.Bvector.
From Stdlib Require Export Vectors.Fin.
From Stdlib Require Export Vectors.FinFun.
From Stdlib Require Export Vectors.Vector.
From Stdlib Require Export Vectors.VectorDef.
From Stdlib Require Export Vectors.VectorEq.
From Stdlib Require Export Vectors.VectorSpec.
From Stdlib Require Export Wellfounded.Disjoint_Union.
From Stdlib Require Export Wellfounded.Inclusion.
From Stdlib Require Export Wellfounded.Inverse_Image.
From Stdlib Require Export Wellfounded.Lexicographic_Exponentiation.
From Stdlib Require Export Wellfounded.Lexicographic_Product.
From Stdlib Require Export Wellfounded.List_Extension.
From Stdlib Require Export Wellfounded.Transitive_Closure.
From Stdlib Require Export Wellfounded.Union.
From Stdlib Require Export Wellfounded.Well_Ordering.
From Stdlib Require Export Wellfounded.Wellfounded.
From Stdlib Require Export ZArith.BinInt.
From Stdlib Require Export ZArith.BinIntDef.
From Stdlib Require Export ZArith.Int.
From Stdlib Require Export ZArith.Wf_Z.
From Stdlib Require Export ZArith.ZArith.
From Stdlib Require Export ZArith.ZArith_base.
From Stdlib Require Export ZArith.ZArith_dec.
From Stdlib Require Export ZArith.ZModOffset.
From Stdlib Require Export ZArith.ZNsatz.
From Stdlib Require Export ZArith.Zabs.
From Stdlib Require Export ZArith.Zbitwise.
From Stdlib Require Export ZArith.Zbool.
From Stdlib Require Export ZArith.Zcompare.
From Stdlib Require Export ZArith.Zcomplements.
From Stdlib Require Export ZArith.Zcong.
From Stdlib Require Export ZArith.Zdiv.
From Stdlib Require Export ZArith.Zdiv_facts.
From Stdlib Require Export ZArith.Zdivisibility.
From Stdlib Require Export ZArith.Zeuclid.
From Stdlib Require Export ZArith.Zeven.
From Stdlib Require Export ZArith.Zgcd_alt.
From Stdlib Require Export ZArith.Zhints.
From Stdlib Require Export ZArith.Zmax.
From Stdlib Require Export ZArith.Zmin.
From Stdlib Require Export ZArith.Zminmax.
From Stdlib Require Export ZArith.Zmisc.
From Stdlib Require Export ZArith.Znat.
From Stdlib Require Export ZArith.Znumtheory.
From Stdlib Require Export ZArith.Zorder.
From Stdlib Require Export ZArith.Zpow_alt.
From Stdlib Require Export ZArith.Zpow_def.
From Stdlib Require Export ZArith.Zpow_facts.
From Stdlib Require Export ZArith.Zpower.
From Stdlib Require Export ZArith.Zquot.
From Stdlib Require Export ZArith.Zwf.
From Stdlib Require Export ZArith.auxiliary.
From Stdlib Require Export Zmod.Bits.
From Stdlib Require Export Zmod.Zmod.
From Stdlib Require Export Zmod.ZmodBase.
From Stdlib Require Export Zmod.ZmodDef.
From Stdlib Require Export Zmod.ZmodInv.
From Stdlib Require Export Zmod.ZmodNsatz.
From Stdlib Require Export Zmod.Zstar.
From Stdlib Require Export Zmod.ZstarBase.
From Stdlib Require Export Zmod.ZstarDef.
From Stdlib Require Export btauto.Algebra.
From Stdlib Require Export btauto.Btauto.
From Stdlib Require Export btauto.Reflect.
From Stdlib Require Export derive.Derive.
From Stdlib Require Export extraction.ExtrHaskellBasic.
From Stdlib Require Export extraction.ExtrHaskellNatInt.
From Stdlib Require Export extraction.ExtrHaskellNatInteger.
From Stdlib Require Export extraction.ExtrHaskellNatNum.
From Stdlib Require Export extraction.ExtrHaskellString.
From Stdlib Require Export extraction.ExtrHaskellZInt.
From Stdlib Require Export extraction.ExtrHaskellZInteger.
From Stdlib Require Export extraction.ExtrHaskellZNum.
From Stdlib Require Export extraction.ExtrOCamlFloats.
From Stdlib Require Export extraction.ExtrOCamlInt63.
From Stdlib Require Export extraction.ExtrOCamlPArray.
From Stdlib Require Export extraction.ExtrOCamlPString.
From Stdlib Require Export extraction.ExtrOcamlBasic.
From Stdlib Require Export extraction.ExtrOcamlChar.
From Stdlib Require Export extraction.ExtrOcamlIntConv.
From Stdlib Require Export extraction.ExtrOcamlNatBigInt.
From Stdlib Require Export extraction.ExtrOcamlNatInt.
From Stdlib Require Export extraction.ExtrOcamlNativeString.
From Stdlib Require Export extraction.ExtrOcamlString.
From Stdlib Require Export extraction.ExtrOcamlZBigInt.
From Stdlib Require Export extraction.ExtrOcamlZInt.
From Stdlib Require Export extraction.Extraction.
From Stdlib Require Export funind.FunInd.
From Stdlib Require Export funind.Recdef.
From Stdlib Require Export micromega.DeclConstant.
From Stdlib Require Export micromega.DeclConstantZ.
From Stdlib Require Export micromega.Env.
From Stdlib Require Export micromega.EnvRing.
From Stdlib Require Export micromega.Fourier.
From Stdlib Require Export micromega.Fourier_util.
From Stdlib Require Export micromega.Lia.
From Stdlib Require Export micromega.Lqa.
From Stdlib Require Export micromega.Lra.
From Stdlib Require Export micromega.OrderedRing.
From Stdlib Require Export micromega.Psatz.
From Stdlib Require Export micromega.QMicromega.
From Stdlib Require Export micromega.RMicromega.
From Stdlib Require Export micromega.Refl.
From Stdlib Require Export micromega.RingMicromega.
From Stdlib Require Export micromega.SatDivMod.
From Stdlib Require Export micromega.Tauto.
From Stdlib Require Export micromega.VarMap.
From Stdlib Require Export micromega.ZArith_hints.
From Stdlib Require Export micromega.ZCoeff.
From Stdlib Require Export micromega.ZMicromega.
From Stdlib Require Export micromega.Zify.
From Stdlib Require Export micromega.ZifyBool.
From Stdlib Require Export micromega.ZifyClasses.
From Stdlib Require Export micromega.ZifyComparison.
From Stdlib Require Export micromega.ZifyInst.
From Stdlib Require Export micromega.ZifyN.
From Stdlib Require Export micromega.ZifyNat.
From Stdlib Require Export micromega.ZifyPow.
From Stdlib Require Export micromega.ZifySint63.
From Stdlib Require Export micromega.ZifyUint63.
From Stdlib Require Export micromega.Ztac.
From Stdlib Require Export nsatz.ENsatzTactic.
From Stdlib Require Export nsatz.NsatzTactic.
From Stdlib Require Export omega.OmegaLemmas.
From Stdlib Require Export omega.PreOmega.
From Stdlib Require Export rtauto.Bintree.
From Stdlib Require Export rtauto.Rtauto.
From Stdlib Require Export setoid_ring.Algebra_syntax.
From Stdlib Require Export setoid_ring.ArithRing.
From Stdlib Require Export setoid_ring.BinList.
From Stdlib Require Export setoid_ring.Cring.
From Stdlib Require Export setoid_ring.Field.
From Stdlib Require Export setoid_ring.Field_tac.
From Stdlib Require Export setoid_ring.Field_theory.
From Stdlib Require Export setoid_ring.InitialRing.
From Stdlib Require Export setoid_ring.Integral_domain.
From Stdlib Require Export setoid_ring.NArithRing.
From Stdlib Require Export setoid_ring.Ncring.
From Stdlib Require Export setoid_ring.Ncring_initial.
From Stdlib Require Export setoid_ring.Ncring_polynom.
From Stdlib Require Export setoid_ring.Ncring_tac.
From Stdlib Require Export setoid_ring.RealField.
From Stdlib Require Export setoid_ring.Ring.
From Stdlib Require Export setoid_ring.Ring_base.
From Stdlib Require Export setoid_ring.Ring_polynom.
From Stdlib Require Export setoid_ring.Ring_tac.
From Stdlib Require Export setoid_ring.Ring_theory.
From Stdlib Require Export setoid_ring.Rings_Q.
From Stdlib Require Export setoid_ring.Rings_R.
From Stdlib Require Export setoid_ring.Rings_Z.
From Stdlib Require Export setoid_ring.ZArithRing.
From Stdlib Require Export ssr.ssrbool.
From Stdlib Require Export ssr.ssrclasses.
From Stdlib Require Export ssr.ssreflect.
From Stdlib Require Export ssr.ssrfun.
From Stdlib Require Export ssr.ssrsetoid.
From Stdlib Require Export ssr.ssrunder.
From Stdlib Require Export ssrmatching.ssrmatching.

